% Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in the
%      documentation and/or other materials provided with the distribution.
%    * Neither the name of the Vrije Universiteit Brussel nor the
%      names of its contributors may be used to endorse or promote products
%      derived from this software without specific prior written permission.
%
%THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
%ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
%WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
%DISCLAIMED. IN NO EVENT SHALL VRIJE UNIVERSITEIT BRUSSEL BE LIABLE FOR ANY
%DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
%(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES%
%LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
%ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
%(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
%SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

% Robust frequency server, based on example in the book "Erlang Programming"
% Illustrates client/server programming in Erlang
% Illustrates process linking to introduce server robustness

-module(frequency).
-export([start/0, stop/0, allocate/0, deallocate/1]).
-export([init/0, tester/0, test/0]).

%% These are the start functions used to create and
%% initialize the server.
start() ->
  register(frequency, spawn(frequency, init, [])).

%% The client Functions
stop() -> call(stop).
allocate()	-> call(allocate).
deallocate(Freq) -> call({deallocate, Freq}).

%% We hide all message passing and the message
%% protocol in a functional interface.
call(Message) ->
  frequency ! {request, self(), Message}, 
  receive
    {reply, Reply} -> Reply
  end.

reply(Pid, Reply) ->
  Pid ! {reply, Reply}.

% Hard-coded frequencies
all_frequencies() -> [10,11,12,13,14,15].

init() ->
  Frequencies = {all_frequencies(), []},
  process_flag(trap_exit, true),
  loop(Frequencies).

loop(Frequencies) ->
  receive
    {request, Pid, allocate} ->
      {NewFrequencies, Reply} = allocate(Frequencies, Pid),
      reply(Pid, Reply),
      loop(NewFrequencies);
    {request, Pid, {deallocate, Freq}} ->
      NewFrequencies = deallocate(Frequencies, Freq), 
      reply(Pid, ok),
      loop(NewFrequencies);
    {request, Pid, stop} ->
      reply(Pid, ok);
    {'EXIT', Pid, _ } ->
      NewFrequencies = auto_deallocate(Frequencies, Pid),
      loop(NewFrequencies)
  end.
  
allocate({[], Allocated}, _Pid) ->
  {{[], Allocated}, {error, no_frequency}};
allocate({[Freq|Free], Allocated}, Pid) ->
  link(Pid),
  {{Free, [{Freq, Pid}|Allocated]}, {ok, Freq}}.

deallocate({Free, Allocated}, Freq) ->
  NewAllocated=lists:keydelete(Freq, 1, Allocated),
  {[Freq|Free], NewAllocated}.
  
% Free = [ Freq ]
% Allocated = [ { Freq, Pid } ]
auto_deallocate({Free, Allocated}, Pid) ->
  case lists:keyfind(Pid, 2, Allocated) of
    {Freq, Pid} ->
      Allocated1 = lists:keydelete(Pid, 2, Allocated),
      auto_deallocate({[Freq|Free], Allocated1}, Pid);
  false ->
    unlink(Pid),
    {Free, Allocated}
end.

expect(Test, Expected, Actual) ->
  case Expected =:= Actual of
    true ->
      io:format("[pass] ~w~n", [Test]);
    false ->
      io:format("[fail] ~w expected:~w got:~w~n", [Test,Expected,Actual])
  end.

test() ->
  spawn_link(frequency, tester, []),
  process_flag(trap_exit, true),
  receive {'EXIT',_Pid,_Reason} ->
    expect(alloc8, {ok, 10}, frequency:allocate()), % all frequencies freed
    expect(stop, ok, frequency:stop())
  end.
  
tester() ->
  expect(start, true, frequency:start()),
  expect(alloc1, {ok,10}, frequency:allocate()),
  expect(alloc2, {ok,11}, frequency:allocate()),
  expect(alloc3, {ok,12}, frequency:allocate()),
  expect(alloc4, {ok,13}, frequency:allocate()),
  expect(alloc5, {ok,14}, frequency:allocate()),
  expect(alloc6, {ok,15}, frequency:allocate()),
  expect(fail_alloc1, {error,no_frequency}, frequency:allocate()),
  expect(dealloc1, ok, frequency:deallocate(11)),
  expect(alloc7, {ok,11}, frequency:allocate()),
  expect(fail_alloc2, {error,no_frequency}, frequency:allocate()).